@@ -3,5 +3,5 @@ from django.utils.translation import gettext_lazy as _
 
 
 class JaccConfig(AppConfig):
-    name = 'jacc'
-    verbose_name = _('Accounting')
+    name = "jacc"
+    verbose_name = _("Accounting")
